import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as animation

from ....utils import bmutils

class mechanism26(bmutils):
    def __init__(self, settings):
        self.settings = settings
        self.fig, self.ax = plt.subplots(figsize=(10, 6))
        self.ani = None

        # Extract parameters from settings
        self.R1 = self.settings['R1']
        self.R2 = self.settings['R2']
        self.R3 = self.settings['R3']
        self.R4 = self.settings['R4']
        self.R_S = self.settings['R_S']
        self.R_A = self.settings['R_A']
        self.R_B = self.settings['R_B']
        self.R_External = self.settings['R_External']
        self.theta_D = self.settings['theta_D']
        # self.theta_forward = self.settings['theta_forward']
        self.previous_theta = 0
        self.instances = settings["instances"]
        self.phase_6 = settings["phase_6"]
        self.theta_rack = 0
        # self.theta_limits = settings["theta_limits"]
        # self.perimeter_theta = settings["perimeter_theta"]
        self.setup_plot()

    def setup_plot(self):
        self.ax.set_title(f'Mechanism 26')
        self.ax.set_xlim(-self.R_B*2, self.R_B*2)
        self.ax.set_ylim(-self.R_B*2, self.R_B*2)
        self.ax.axhline(0, color='black', linewidth=0.5, linestyle='--')
        # self.ax.grid()
        self.ax.axis("off")
        self.ax.set_aspect('equal')
        super().watermark(self.ax)
    
    def plot(self, theta = 0):    
        self.setup_plot()
        # the racks (fixed internal gears)
        x_A, y_A = self.create_circle_section(self.R_A)
        self.ax.plot(x_A, y_A, color='black', linestyle='-.')
        self.ax.plot([0], [0], 'ro') 

        d_theta_1 = theta - self.previous_theta
        self.previous_theta = theta


        x1, y1 = self.create_circle_section(self.R1, phase=theta)
        self.ax.plot(x1, y1, color='red', linestyle='dotted')
        x1, y1 = self.create_circle_section(self.R1, phase=theta, end_angle=0.2*np.pi) # marker (solid section)
        self.ax.plot(x1, y1, color='red', linestyle='-', linewidth=5)

        
        R1 = self.R1
        R2 = self.R2
        R3 = self.R3
        R4 = self.R4
        R_S = self.R_S
        R_A = self.R_A
        R_B = self.R_B
        R_External = self.R_External
       


        ## rotate wheel about
        x_B, y_B = self.create_circle_section(self.R_B, phase=self.phase_6+self.theta_rack, end_angle=self.theta_D)
        
        for index, instance in enumerate(self.instances):
            self.__dict__.update(instance)
            self.theta += d_theta_1
            R2_center =   [R1+R2  , 0]
            R_intersect = [R1+2*R2, 0]

            centerx, centery = self.rotate_points(self.phase_6, [R2_center[0]],[R2_center[1]], [0,0])
            R2_center = [centerx[0],centery[0]]
            centerx, centery = self.rotate_points(self.phase_6, [R_intersect[0]],[R_intersect[1]], [0,0])
            R_intersect = [centerx[0],centery[0]]
            
            

            instance_theta = self.theta #+ d_theta_1
            if instance_theta != 0:
                direction = instance_theta/abs(instance_theta)
                self.previous_direction = direction
            else:
                direction = self.previous_direction
            self.previous_direction = direction
            instance_theta = direction * (abs(instance_theta) % (2*np.pi))


            d_theta_2 = - R1/R2 * self.theta # self.theta is theta of R1
            # d_theta_A = d_theta_2 * R2 / R_A
            d_theta_A = -d_theta_2 * R3 / R_A #+ self.theta_D/2
            # print(index, d_theta_A)
            # centerx, centery = self.rotate_points(instance_theta, [R2_center[0]],[R2_center[1]], [0,0])
            centerx, centery = self.rotate_points(d_theta_A, [R2_center[0]],[R2_center[1]], [0,0]) # it is starting at the wrong
            R2_center = [centerx[0],centery[0]]
            self.ax.plot([R2_center[0]], [R2_center[1]], 'ro')

            centerx, centery = self.rotate_points(d_theta_A, [R_intersect[0]],[R_intersect[1]], [0,0])
            R_intersect = [centerx[0],centery[0]]
            # self.ax.plot([R_intersect[0]], [R_intersect[1]], 'ro')

           
            x2, y2 = self.create_circle_section(self.R2, phase=d_theta_2)
            x3, y3 = self.create_circle_section(self.R3, phase=d_theta_2)
            self.ax.plot(x2+R2_center[0], y2+R2_center[1], color='red', linestyle='dotted')
            self.ax.plot(x3+R2_center[0], y3+R2_center[1], color='black', linestyle='dotted')

            


            if d_theta_A != 0:
                direction = d_theta_A/abs(d_theta_A)
            else:
                direction = 1
            d_theta_A = direction * (abs(d_theta_A) % 2*np.pi)
            theta_rack_start = self.theta_rack
            

            if theta_rack_start != 0:
                direction = theta_rack_start/abs(theta_rack_start)
            else:
                direction = 1
            theta_rack_start = direction * (abs(theta_rack_start) % 2*np.pi)
            theta_rack_stop = theta_rack_start + self.theta_D

            theta_rack_start = self.theta_rack
            theta_rack_start = theta_rack_start if theta_rack_start > 0 else theta_rack_start + 2* np.pi
            theta_rack_start = theta_rack_start if theta_rack_start < 2* np.pi else theta_rack_start - 2* np.pi
            


            # gear_is_active = (theta_rack_start <= d_theta_A and d_theta_A < theta_rack_stop)
            gear_is_active = self.is_point_on_circle(R_intersect, x_B, y_B)
            
            if gear_is_active:                
                _d_theta_2 = R1/R2 * d_theta_1
                if instance["type"] == 0:
                    active_rack_r = R_B
                    active_planet_r = R2
                    # print("active", index, instance["type"])
                else:
                    active_planet_r = R4
                    active_rack_r = R_S
                _d_theta_A = _d_theta_2 * R3/R_A - _d_theta_2 * active_planet_r/active_rack_r
                self.theta_rack += _d_theta_A

            if instance["type"] == 0:
                self.ax.plot(x2+R2_center[0], y2+R2_center[1], color='green', linestyle=':')
                x2, y2 = self.create_circle_section(self.R2, phase=d_theta_2, end_angle=0.1*np.pi)
                self.ax.plot(x2+R2_center[0], y2+R2_center[1], color='green', linestyle='-', linewidth=4)
            else:           
                x4, y4 = self.create_circle_section(self.R4, phase=d_theta_2)
                self.ax.plot(x4+R2_center[0], y4+R2_center[1], color='blue', linestyle='--')
                x2, y2 = self.create_circle_section(self.R2, phase=d_theta_2, end_angle=0.1*np.pi)
                self.ax.plot(x2+R2_center[0], y2+R2_center[1], color='red', linestyle='-', linewidth=4)

            
            
            for key, value in self.__dict__.items():
                if key in instance:  # Check if the key exists in the instance
                    instance[key] = value  # Update the dictionary
                elif hasattr(instance, '__dict__') and key in vars(instance):
                    setattr(instance, key, value)
        #     pass
        x_S, y_S = self.create_circle_section(self.R_S, phase=self.phase_6+self.theta_rack, end_angle=self.theta_D) # __check__ need to calculate end_angle depending on number of arms
        self.ax.plot(x_S, y_S, color='blue',linestyle='--')
        # x_B, y_B = self.create_circle_section(self.R_B, phase=self.phase_6+self.theta_rack, end_angle=self.theta_D) # __check__ need to calculate end_angle depending on number of arms
        self.ax.plot(x_B, y_B, color='green',linestyle='--')
        x_External, y_External = self.create_circle_section(R_External, phase=self.phase_6+self.theta_rack)
        self.ax.plot(x_External, y_External, color='black',linestyle='--')
        
        ''' Markers '''
        x_External, y_External = self.create_circle_section(R_External, phase=self.phase_6+self.theta_rack, end_angle=0.1*np.pi)
        self.ax.plot(x_External, y_External, color='black', linestyle='-', linewidth=5)

        x_External, y_External = self.create_circle_section(self.R_External, phase=self.phase_6+self.theta_rack, end_angle=self.theta_D)
        # Determine the start and end points of the segments
        start_S = (x_S[0], y_S[0])  # Start of the first segment
        end_S = (x_S[-1], y_S[-1])  # End of the first segment
        start_B = (x_External[0], y_External[0])  # Start of the second segment
        end_B = (x_External[-1], y_External[-1])  # End of the second segment

        # Draw lines between the points
        self.ax.plot([start_S[0], start_B[0]], [start_S[1], start_B[1]], color='black') 
        self.ax.plot([end_S[0], end_B[0]], [end_S[1], end_B[1]], color='black')  

        return self
    
    def update(self, interval=100, frames=100):
        speed = self.settings["speed"]
        len = self.settings["animation_rounds"]
        displacements = np.linspace(0, len, int(250/speed))
        # displacements = np.concatenate([displacements, displacements[::-1]])  # Reverse motion
        def animate(displacement):
            self.ax.clear()
            self.plot(displacement)

        
        self.setup_plot()
        self.ani = animation.FuncAnimation(self.fig, lambda displacement: animate(displacement), frames=displacements, interval=50, repeat=True)
        return self
        

    @staticmethod
    # return_speed: how many times return is faster than forward
    # frequency: number of oscillations of output per revolution of planet. Should be greater than one
    def Settings(R1=5,R2=20, return_speed=2, frequency = 1, animation_rounds=1, speed=1):
        num_hands = frequency *2
        theta_segment = 2*np.pi/num_hands # angle of rack segment used
        theta_D = theta_segment # theta_disk

        k = return_speed

        R_B = R1 + 2 * R2 # R_Big
        theta_2_R = theta_D * R_B / R2  # Disk 2 rotates through theta_2_R to cover distance of theta_D * R_B(entire length of rack)

        theta_D_R  = (1/(k+1) ) * theta_D
        theta_D_F = 2 * theta_D - theta_D_R

        theta_3_R = theta_2_R
        c = theta_D_R / theta_3_R
        R3 = (c * R1 + c * R2)/(1-c)
        R_A = R1 + R2 + R3  # R_Annulus
        

        theta_3_F = theta_D_F * R_A / R3
        theta_4_F = theta_3_F
        c = theta_D / theta_4_F

        R4 = c*(R1 + R2) / (1-c)
        R_S = R1 + R2 + R4

        R_External = R_B * 1.2

        theta_3_perimeter = 2*np.pi * R_A / R3
        s_2_perimeter = theta_3_perimeter * R2
        theta_1_perimeter = s_2_perimeter / R1
        perimeter_theta = theta_1_perimeter
        animation_rounds = int(animation_rounds * perimeter_theta)
        num_instances = num_hands
        theta_points = [(perimeter_theta/num_instances) * i for i in range(num_instances)]
        instances = [{"theta":theta, "type":index%2} for index, theta in enumerate(theta_points)]
        for index, instance in enumerate(instances):
            instance["previous_direction"] = 1

        
        # phase_6 = - theta_D_F/2 - theta_D_R/2 - np.pi/2
        phase_6 = - theta_D_R/2 - np.pi/2
        return {
            "R1": R1,
            "R2": R2,
            "R3": R3,
            "R4": R4,
            "R_S": R_S,
            "R_A": R_A,
            "R_B": R_B,
            "R_External": R_External,
            "theta_D": theta_D,
            # "theta_forward":theta_forward,
            "animation_rounds":animation_rounds,
            "speed":speed,
            "perimeter_theta":perimeter_theta,
            # "theta_limits":theta_limits,
            "instances":instances,
            "phase_6":phase_6,
            "frequency":frequency,
        }
    @staticmethod
    def info():
        description = (
            "Mechanism Description:\n"
            "-----------------------\n"
            "Continuous rotary to reciprocating rotary and/or linear motion with possible configuration for quick return.\n"
            "**Applications**:\n"
            "This type of mechanism is intended to be used in the development of clamp-on self-actuated long-range linear motion mechanisms.\n"
        )
        return description



import numpy as np

thetaRotations = {
    # add more values < -0.25
    "-0.1667": 38, # todo: check this value
    "-0.2":0,
    "-0.25":90,
    "-0.3333":45,
    "-0.5":90,
    "-2": 90,
    "-3":45,
    "-4":90,
    "-5":0,
    "-6":38, # todo: check this value
    # todo: add more values < -6
}

shapeParams={
    "1": {
        "name": "Straight Line",
        "options":[
            {
               "len_factor": 1,
                "speed_factor": -1,
            },
            {
                "len_factor": 4, # todo: check this value
                "speed_factor": -0.5,
                # "thetaMax": 180
                "thetaMax": 180-50
            },
        ]
    },
    "3":{
        "name": "Triangle",
        "options":[
            {
                "len_factor": 0.25,
                "speed_factor": -2
            },
            {
                "len_factor": 4, # todo: check this value
                "speed_factor": -0.5
            },
        ]
    },
    "4": {
        "name": "Square",
        "options":[
            {
                "len_factor": 0.135, # todo: check this value
                "speed_factor": -3
            },
            {
                "len_factor": 1/0.135,
                "speed_factor": -1/3
            },
        ]
    },
    # todo: add more shapes
}

class helpers:
    def __init__(self):
        pass

    @staticmethod
    def shapeParams(numSides):
        return shapeParams[f"{numSides}"] if f"{numSides}" in shapeParams else None
    @staticmethod
    def create2linksForBar(numBars=1, speed_factor=None, len_factor=None, numLinks=None, numSides=None, shapeOption=0, theta0=0):
        thetaMax = None
        if numSides is not None:
            shapeParam = helpers.shapeParams(numSides)
            if shapeParam is None:
                # throw error
                raise ValueError(f"Shape with {numSides} sides not found")
            len_factor = shapeParam["options"][shapeOption]["len_factor"]
            speed_factor = shapeParam["options"][shapeOption]["speed_factor"]
            thetaMax = shapeParam["options"][shapeOption]["thetaMax"] if "thetaMax" in shapeParam["options"][shapeOption] else None
        alphaBetLetters = [chr(i) for i in range(65, 65+numBars*2)]
        oddNumbers = [i for i in range(1, numBars*2+1, 2)]
        if thetaMax is None:
            if abs(speed_factor) < 1:
                thetaMax = 360/np.abs(speed_factor)
            else:
                thetaMax = 360
        
        links = []
        angle = thetaMax/numBars
        for i in range(1, numBars+1):
            link1 = {
                "points": {
                    f"O{i}": 0, f"A{i}":1
                },
                "speed": 1,
                "equation": "np.cos(x), np.sin(x)",
                "origin": "{0,0}",
                "theta0": angle*(i-1),
            }
            link2 = {
                "points": {
                    f"A{i}":0, f"B{i}":1
                },
                "len_factor": len_factor,
                "speed_factor": speed_factor,
                "speed_source": f"link1.speed",
                "equation": "np.cos(speed_factor * speed_source * x), np.sin(speed_factor * speed_source * x)",
                "origin": f"link{oddNumbers[i-1]}.A{i}",
                "theta0": angle*(i-1)+theta0,
                "output": f"B{i}"
            }
            links.append(link1)
            links.append(link2)
        linksTemplate = {}
        if numLinks is None:
            numLinks = numBars * 2
        for i in range(1, numLinks+1):
            linksTemplate[f"link{i}"] =links[i-1].copy()

        thetaRotationIndex = f"{speed_factor}" if speed_factor <-1 else f"{speed_factor:.4f}"
        while thetaRotationIndex[-1] == "0":
            thetaRotationIndex = thetaRotationIndex[:-1]
        if f"{thetaRotationIndex}" in thetaRotations:
            thetaRotation = thetaRotations[f"{thetaRotationIndex}"]
        else:
            thetaRotation = 0

        
        return linksTemplate, thetaMax, thetaRotation


from ._1 import *
from ._2 import *
from ._13 import *
from ._14 import *
from ._22 import *
from ._23 import *
from ._25 import *
from ._26 import *
from ._27 import *
from ._28 import *
from ._29 import *
from ._30 import *
from ._31 import *
from ._32 import *
from ._33 import *


import matplotlib.pyplot as plt
import numpy as np
from matplotlib.animation import FuncAnimation

class ULMW():
    def __init__(self, settings):
        self.head_start = settings.head_start
        self.r1 = settings.r1
        self.r2 = settings.r2
        self.r3 = settings.r3
        self.rack_length = settings.rack_length
        self.ULMWDia = settings.ULMWDia
        self.ULMWPerimeter = settings.ULMWPerimeter
        self.fig = settings.fig
        self.ax = settings.ax

        self.endPerimeter = np.pi * self.ULMWDia / 2
        self.gearArmCenterDistance = self.r1 + self.r2
        self.gear1Perimenter = np.pi * self.r1*2
        self.gear2Perimenter = np.pi * self.r2*2
        self.gear3Perimenter = np.pi * self.r3*2     
        
        if self.ax is not None:
            self.centerLine, = self.ax.plot([], [], f"{settings.colours[0]}--")
            self.side_right, = self.ax.plot([], [], f"{settings.colours[0]}-")
            self.side_left, = self.ax.plot([], [], f"{settings.colours[0]}-")
            self.semi_circle_top, = self.ax.plot([], [], f"{settings.colours[0]}-")
            self.semi_circle_bottom, = self.ax.plot([], [], f"{settings.colours[0]}-")
            self.gear_arm_line, = self.ax.plot([], [], f"{settings.colours[1]}-")  # Line representing gear arm

            self.gear1, = self.ax.plot([], [], f"{settings.colours[2]}-")
            self.gear2, = self.ax.plot([], [], f"{settings.colours[0]}-")
            self.gear3, = self.ax.plot([], [], f"{settings.colours[3]}-")

        

        self.checkPoints = [
            self.rack_length / 2,  # Quadrant 1
            self.rack_length / 2 + self.endPerimeter,  # Quadrant -1
            self.rack_length / 2 + self.endPerimeter + self.rack_length / 2,  # Quadrant 2
            self.rack_length / 2 + self.endPerimeter + self.rack_length,  # Quadrant 3
            self.rack_length / 2 + self.endPerimeter + self.rack_length + self.endPerimeter,  # Quadrant -2
            self.rack_length / 2 + self.endPerimeter + self.rack_length + self.endPerimeter + self.rack_length / 2  # Quadrant 4
        ]

        
        self.theta1 = self.theta2 = self.theta3 = 0

    def create_gear(self, r, num_teeth=20):
        theta = np.linspace(0, 2 * np.pi, num_teeth, endpoint=False)
        x = r * np.cos(theta)
        y = r * np.sin(theta)
        return x, y

    def setPosition(self, position):
        position+=self.head_start
        # Calculate the quadrant and remaining position
        gear1Rem=position %(self.gear1Perimenter)
        self.theta1 = -360* gear1Rem/self.gear1Perimenter

        gear2Rem=position %(self.gear2Perimenter)
        self.theta2 = 360* gear2Rem/self.gear2Perimenter

        gear3Rem=position %(self.gear3Perimenter)
        self.theta3 = 360* gear3Rem/self.gear3Perimenter

        remPosition = position % self.ULMWPerimeter
        position = remPosition
        for i, point in enumerate(self.checkPoints):
            if remPosition < point:
                rackQuadrant = i
                break

        armAngle = 0  # degrees

        if rackQuadrant == 0:  # [0]
            rackAssemblyCenterPosition = -position  # a negative value
        elif rackQuadrant == 1:  # [1] # Dwell
            armAngle = 180 - 180 * (self.checkPoints[1] - position) / (self.checkPoints[1] - self.checkPoints[0])
            rackAssemblyCenterPosition = -(self.rack_length / 2)  # negative value
        elif rackQuadrant == 2:  # [2]
            armAngle = 180
            rackAssemblyCenterPosition = -(self.checkPoints[2] - position)  # negative value
        elif rackQuadrant == 3:  # [0]
            armAngle = 180
            rackAssemblyCenterPosition = (self.rack_length / 2) - (self.checkPoints[3] - position)  # +ve value
        elif rackQuadrant == 4:  # Dwell
            armAngle = 360 - 180 * (self.checkPoints[4] - position) / (self.checkPoints[4] - self.checkPoints[3])
            rackAssemblyCenterPosition = self.rack_length / 2
        else:  # rackQuadrant == 5
            rackAssemblyCenterPosition = self.checkPoints[5] - position  # +ve value

        return rackAssemblyCenterPosition, armAngle
    
    def armAnglePoints(self):
        arm_angles = []
        for position in range(int(self.ULMWPerimeter)):
            _, arm_angle = self.setPosition(position)
            if arm_angle > 180:
                arm_angle = 180 - (arm_angle-180)
            arm_angles.append(arm_angle)
        return arm_angles

    def plot_arm_angles(self):
        import matplotlib.pyplot as plt

        arm_angles = self.armAnglePoints()
        positions = range(int(self.ULMWPerimeter))

        plt.figure(figsize=(12, 6))
        plt.plot(positions, arm_angles)
        plt.title('Arm Angle vs Position')
        plt.xlabel('Position')
        plt.ylabel('Arm Angle (degrees)')
        plt.grid(True)
        plt.show()

    @staticmethod
    def plot_multiple_arm_angles(instances, labels=None, save=None):
        # plt.figure(figsize=(12, 6))
        
        for i, instance in enumerate(instances):
            arm_angles = instance.armAnglePoints()
            positions = range(int(instance.ULMWPerimeter))
            
            label = labels[i] if labels and i < len(labels) else f'Instance {i+1}'
            plt.plot(positions, arm_angles, label=label)
        
        plt.suptitle('Arm Location(angle)', fontsize=14)
        plt.title(f"r_ULMW={instances[0].ULMWDia/2}, length={instances[0].rack_length}", fontsize=10)
        plt.xlabel('Arm Distance')
        plt.ylabel('Arm Angle (degrees)')
        plt.grid(True)
        plt.legend()
        
        if save:
            plt.savefig(save)
        plt.show()
    
    @staticmethod
    def plot_dynamic_arm_angles(num_instances, ULMWRadius, rack_length, labels=None, save=None):
        instances = []
        r1 = r2 = r3 = ULMWRadius
        
        # Create instances
        for i in range(num_instances):
            settings = Settings(rack_length, r1, r2, r3, None, None, numULMWs=num_instances, num=i)
            instance = ULMW(settings)
            instances.append(instance)

        # Plot
        plt.figure(figsize=(12, 6))
        
        for i, instance in enumerate(instances):
            arm_angles = instance.armAnglePoints()
            positions = range(int(instance.ULMWPerimeter))
            
            label = labels[i] if labels and i < len(labels) else f'Instance {i+1}'
            plt.plot(positions, arm_angles, label=label)
        
        plt.suptitle('Arm Location (angle)', fontsize=14)
        plt.title(f"r_ULMW={ULMWRadius}, rack_length={rack_length}", fontsize=10)
        plt.xlabel('Arm Distance')
        plt.ylabel('Arm Angle (degrees)')
        plt.grid(True)
        plt.legend()
        
        if save:
            plt.savefig(save)
        plt.show()
    
    @staticmethod
    def plot_dynamic_arm_angles1(num_instances, ULMWRadius, rack_length, labels=None, zero_threshold=0.1,save=None):
        instances = []
        r1 = r2 = r3 = ULMWRadius
        
        # Create instances
        for i in range(num_instances):
            settings = Settings(rack_length, r1, r2, r3, None, None, numULMWs=num_instances, num=i)
            instance = ULMW(settings)
            instances.append(instance)

        # Collect data
        all_arm_angles = []
        for instance in instances:
            arm_angles = instance.armAnglePoints()
            all_arm_angles.append(arm_angles)

        # Convert to numpy array for easier manipulation
        all_arm_angles = np.array(all_arm_angles)

        # Count instances with arm angle close to 0 for each position
        zero_count = np.sum(np.abs(all_arm_angles) < zero_threshold, axis=0)

        # Plot
        fig, (ax1, ax2) = plt.subplots(2, 1, figsize=(12, 10), sharex=True)
        
        # Plot arm angles
        for i, arm_angles in enumerate(all_arm_angles):
            positions = range(int(instances[0].ULMWPerimeter))
            label = labels[i] if labels and i < len(labels) else f'Instance {i+1}'
            ax1.plot(positions, arm_angles, label=label)
        
        ax1.set_ylabel('Arm Angle (degrees)')
        ax1.grid(True)
        ax1.legend()

        # Plot count of instances with arm angle = 0
        ax2.plot(positions, zero_count, color='red', linewidth=2)
        ax2.set_xlabel('Arm Distance')
        ax2.set_ylabel('Count of Instances\nwith Arm Angle ≈ 0')
        ax2.grid(True)

        plt.suptitle('Arm Location Analysis', fontsize=14)
        plt.title(f"r_ULMW={ULMWRadius}, rack_length={rack_length}", fontsize=10)
        plt.tight_layout()
        if save:
            plt.savefig(save)
        plt.show()

    def init_plot(self):
        # Initialize the plot limits and labels
        self.ax.set_xlim(-2 * self.ULMWDia, 2 * self.ULMWDia)
        self.ax.set_ylim(-30, 30)
        plt.title("Position Animation")
        plt.xlabel('Position')
        plt.ylabel('Y-axis')
        plt.grid(True)
        return self.centerLine, self.side_right, self.side_left, self.semi_circle_top, self.semi_circle_bottom, self.gear_arm_line, self.gear1, self.gear2, self.gear3

    def create_semi_circle(self, offset_y, direction=1):
        radius = self.ULMWDia / 2
        theta = np.linspace(0, np.pi * direction, 100)
        x = radius * np.cos(theta)
        y = radius * np.sin(theta) + offset_y
        return x, y

    def update_plot(self, position):
        # Update the vertical line position
        rackAssemblyCenterPosition, armAngle = self.setPosition(position)
        self.centerLine.set_data([0, 0], [rackAssemblyCenterPosition - self.rack_length / 2,
                                          rackAssemblyCenterPosition + self.rack_length / 2])
        self.side_right.set_data([self.ULMWDia / 2, self.ULMWDia / 2], [rackAssemblyCenterPosition - self.rack_length / 2,
                                                                        rackAssemblyCenterPosition + self.rack_length / 2])
        self.side_left.set_data([-self.ULMWDia / 2, -self.ULMWDia / 2], [rackAssemblyCenterPosition - self.rack_length / 2,
                                                                         rackAssemblyCenterPosition + self.rack_length / 2])

        x_top, y_top = self.create_semi_circle(rackAssemblyCenterPosition + self.rack_length / 2)
        x_bottom, y_bottom = self.create_semi_circle(rackAssemblyCenterPosition - self.rack_length / 2, -1)

        self.semi_circle_top.set_data(x_top, y_top)
        self.semi_circle_bottom.set_data(x_bottom, y_bottom)

        # Calculate the end coordinates of the gear arm line
        end_x = self.gearArmCenterDistance * np.cos(np.radians(armAngle))
        end_y = self.gearArmCenterDistance * np.sin(np.radians(armAngle))
        self.gear_arm_line.set_data([0, end_x], [0, end_y])

        x1, y1 = self.create_gear(self.r1)
        x2, y2 = self.create_gear(self.r2)
        x3, y3 = self.create_gear(self.r3)

        theta1 = np.radians(self.theta1)
        theta2 = np.radians(self.theta2)
        theta3 = np.radians(self.theta3)
        # if armAngle not in [0, 180]:
        theta3 += np.radians(armAngle)
        theta2 = theta3

        # Apply rotations
        x1r = x1 * np.cos(theta1) - y1 * np.sin(theta1)
        y1r = x1 * np.sin(theta1) + y1 * np.cos(theta1)

        x2r = x2 * np.cos(theta2) - y2 * np.sin(theta2)
        y2r = x2 * np.sin(theta2) + y2 * np.cos(theta2)

        x3r = x3 * np.cos(theta3) - y3 * np.sin(theta3)
        y3r = x3 * np.sin(theta3) + y3 * np.cos(theta3)

        # Update data for the gears
        self.gear1.set_data(x1r, y1r)
        self.gear2.set_data(x2r + end_x, y2r + end_y)
        self.gear3.set_data(x3r + end_x, y3r + end_y)

        return self.centerLine, self.side_right, self.side_left, self.semi_circle_top, self.semi_circle_bottom, self.gear_arm_line, self.gear1, self.gear2, self.gear3

    def animatePosition(self):
        # Create animation
        frames_forward = np.arange(0, 101)  # Forward frames from 0 to 100
        frames_backward = np.arange(100, -1,-1)  # Backward frames from 100 to 0
        # Concatenate the forward and backward frames
        frames = np.concatenate((frames_forward, frames_backward))
        anim = FuncAnimation(self.fig, self.update_plot, frames=frames,
                             init_func=self.init_plot, blit=True)
        plt.show()

    @staticmethod
    def dMax(working_length, rack_length, num_sets): # maximum ULMW Diameter
        dMax = (working_length * num_sets - 2 * rack_length)/np.pi
        return dMax
    
    @staticmethod
    def lMin(working_length, d, num_sets): # minimum rack_length
        rack_length = (working_length * num_sets - np.pi *d)/2
        return rack_length


# Define a sample Settings class
class Settings:
    def __init__(self, rack_length, r1, r2, r3, fig, ax, head_start=0, numULMWs=1, num=0):
        self.rack_length = rack_length
        self.r1 = r1
        self.r2 = r2
        self.r3 = r3
        
        self.fig = fig
        self.ax = ax

        self.ULMWDia = (r1 + r2 - r3) * 2
        self.ULMWPerimeter = 2 * rack_length + np.pi * self.ULMWDia
        #head_start
        self.head_start = head_start
        if numULMWs > 1:
            self.head_start = (num/numULMWs)*self.ULMWPerimeter

        if ax is not None:
            ax.set_xlim(-2 * self.ULMWDia, 2 * self.ULMWDia)
            len = 2*self.rack_length+self.ULMWDia
            ax.set_ylim(-len/2, len/2)

        # Define list of colors using single-character codes
        colours = [
            ['k', 'r', 'g', 'b'],  # Set 1
            ['c', 'm', 'y', 'k'],  # Set 2
            ['m', 'y', 'b', 'g'],  # Set 3
            ['r', 'g', 'b', 'c'],  # Set 4
            ['y', 'b', 'g', 'k'],  # Set 5
            ['k', 'c', 'r', 'm'],  # Set 6
            ['g', 'y', 'c', 'b'],  # Set 7
            ['b', 'k', 'y', 'r'],  # Set 8
            ['m', 'c', 'k', 'g'],  # Set 9
            ['r', 'y', 'm', 'b']   # Set 10
        ]
        
        # Assign colors based on num
        # self.colours = colours[int(num % len(colours))]
        self.colours = colours[num % 10] # problem with len(colours)




import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as animation

class mechanism23:
    def __init__(self, settings):
        self.settings = settings
        self.fig, self.ax = plt.subplots(figsize=(10, 6))
        self.ani = None

        # Extract parameters from settings
        self.R1 = self.settings['R1']
        self.R2 = self.settings['R2']
        self.R3 = self.settings['R3']
        self.R4 = self.settings['R4']
        self.L = self.settings['L']
        self.centers = self.settings['centers']
        self.origin = self.settings['origin']
        self.bar_length = self.L * 2
        self.bar_height = self.R4
        self.bar_height_max_disp = 0.5
        # self.x = 0
        # self.previous_theta = 0
        # self.direction = 1
        # self.previous_direction = 1
        # self.arm_angle = 0
        self.instances = settings["instances"]
        self.setup_plot()


    def setup_plot(self):
        self.ax.set_xlim(self.settings['x_limits'][0], self.settings['x_limits'][1])
        self.ax.set_ylim(-2 * self.R4, 2 * self.R4)
        self.ax.axhline(0, color='black', linewidth=0.5, linestyle='--')
        self.ax.grid()
        self.ax.set_aspect('equal')

    def create_circle_section(self, r, num_points=100, start_angle=0, end_angle=2 * np.pi, phase=0):
        theta = np.linspace(start_angle + phase, end_angle + phase, num_points)
        x = r * np.cos(theta)
        y = r * np.sin(theta)
        return x, y
    
    # Function to find the closest point on a line segment to a point
    def closest_point_on_segment(self, px, py, x1, y1, x2, y2):
        # Vector from x1,y1 to x2,y2
        segment_vec = np.array([x2 - x1, y2 - y1])
        segment_length_sq = np.dot(segment_vec, segment_vec)

        if segment_length_sq == 0:  # x1, y1 and x2, y2 are the same point
            return x1, y1

        # Project point onto the line segment, computing t
        t = np.dot(np.array([px - x1, py - y1]), segment_vec) / segment_length_sq
        t = np.clip(t, 0, 1)  # Clamp t to the segment
        closest_x = x1 + t * segment_vec[0]
        closest_y = y1 + t * segment_vec[1]
        return closest_x, closest_y
    
    # Find the closest point on the cam profile
    def cam_is_active(self, cam_x_values, cam_y_values, midpoint, line, line_floating, perpendicular, perpendicular_floating, cam_distance):
        x_perpendicular,y_perpendicular = perpendicular
        x_line, y_line = line
        x_line_floating, y_line_floating = line_floating
        x_new_line_floating, y_new_line_floating = perpendicular_floating
        midpoint_x, midpoint_y = midpoint
        midpoint_x_o, midpoint_y_o = midpoint
        closest_dist = float('inf')
        closest_point = (None, None)

        centers = self.centers
        if self.x == centers[0] or self.x == centers[-1]:
                rotation_theta = self.arm_angle if self.direction == 1 else self.arm_angle + np.pi
                rotation_matrix = np.array([[np.cos(rotation_theta), -np.sin(rotation_theta)],
                                        [np.sin(rotation_theta), np.cos(rotation_theta)]])
                rotation_center = centers[0] if midpoint_x < 0 else centers[-1]
                midpoint_xs, midpoint_ys = self.rotate_points(rotation_matrix, [midpoint_x], [midpoint_y], rotation_center, 0)
                midpoint_x = midpoint_xs[0]
                midpoint_y = midpoint_ys[0]
                x_line, y_line = self.rotate_points(rotation_matrix, x_line, y_line, rotation_center, 0)
                x_line_floating, y_line_floating = self.rotate_points(rotation_matrix, x_line_floating, y_line_floating, rotation_center, 0)
                x_new_line_floating, y_new_line_floating = self.rotate_points(rotation_matrix, x_new_line_floating, y_new_line_floating, rotation_center, 0)
                x_perpendicular,y_perpendicular = self.rotate_points(rotation_matrix, x_perpendicular,y_perpendicular, rotation_center, 0)
        
        else:
            if self.direction != self.previous_direction:
                rotation_theta = -np.pi #if self.direction == -1 else 0
                rotation_matrix = np.array([[np.cos(rotation_theta), -np.sin(rotation_theta)],
                                 [np.sin(rotation_theta), np.cos(rotation_theta)]])
                rotation_center = self.x # centers[0] if midpoint_x < 0 else centers[-1]
                midpoint_xs, midpoint_ys = self.rotate_points(rotation_matrix, [midpoint_x], [midpoint_y], rotation_center, 0)
                midpoint_x = midpoint_xs[0]
                midpoint_y = midpoint_ys[0]
                x_line, y_line = self.rotate_points(rotation_matrix, x_line, y_line, rotation_center, 0)
                x_line_floating, y_line_floating = self.rotate_points(rotation_matrix, x_line_floating, y_line_floating, rotation_center, 0)
                x_new_line_floating, y_new_line_floating = self.rotate_points(rotation_matrix, x_new_line_floating, y_new_line_floating, rotation_center, 0)
                x_perpendicular,y_perpendicular = self.rotate_points(rotation_matrix, x_perpendicular,y_perpendicular, rotation_center, 0)

        for i in range(len(cam_x_values) - 1):
            x1, y1 = cam_x_values[i], cam_y_values[i]
            x2, y2 = cam_x_values[i + 1], cam_y_values[i + 1]
            closest_x, closest_y = self.closest_point_on_segment(midpoint_x, midpoint_y, x1, y1, x2, y2)
            
            # Calculate distance from the midpoint to the closest point
            dist = np.sqrt((closest_x - midpoint_x) ** 2 + (closest_y - midpoint_y) ** 2)
            
            if dist < closest_dist:
                closest_dist = dist
                closest_point = (closest_x, closest_y)

            if closest_dist > (cam_distance):
                cam_active = False
            else:
                cam_active = True

        return cam_active, [midpoint_x, midpoint_y], closest_point, [x_line, y_line], [x_line_floating, y_line_floating], [x_perpendicular, y_perpendicular], [x_new_line_floating, y_new_line_floating]
    
    def rotate_points(self, rotation_matrix, x_points, y_points, center_x, center_y):
            # Convert lists to numpy arrays
            x_points = np.array(x_points)
            y_points = np.array(y_points)

            # Translate points to the origin
            translated_x = x_points - center_x
            translated_y = y_points - center_y
            
            # Rotate points
            rotated = rotation_matrix @ np.array([translated_x, translated_y])
            
            # Translate points back
            return rotated[0] + center_x, rotated[1] + center_y

    def plot(self, theta = 0):    
        self.ax.set_title(f'Mechanism 23')
        self.ax.set_xlim(self.settings['x_limits'][0], self.settings['x_limits'][1])
        self.ax.set_ylim(-2 * self.R4, 2 * self.R4)
        self.ax.axis("off")
        theta_g = theta
        for index, instance in enumerate(self.instances):    
            self.__dict__.update(instance)
            
            theta = theta_g + self.theta_offset
            theta_R2 = -theta
            x1, y1 = self.create_circle_section(self.R1, phase=theta_R2)
            x2, y2 = self.create_circle_section(self.R2, phase=theta_R2)
            
            
            centers = self.centers
            if index == 1: # plot only once
                for i,center in enumerate(centers): # the compound gears
                    self.ax.plot(x1+center, y1, color='black', linestyle='--')
                    self.ax.plot(x2+center, y2, color='red', linestyle='--')
            
            # the cam profiles
            x3, y3 = self.create_circle_section(self.R3, end_angle=np.pi*0.8, phase=np.pi/2+0.2*np.pi)
            x4, y4 = self.create_circle_section(self.R4, end_angle=np.pi*0.8, phase=np.pi/2)
            x3_b, y3_b = self.create_circle_section(self.R3, end_angle=np.pi*0.8, phase=-np.pi/2)
            x4_b, y4_b = self.create_circle_section(self.R4, end_angle=np.pi*0.8, phase=-np.pi/2+0.2*np.pi)

            d_x = 0

            d_theta = theta - self.previous_theta
            self.previous_theta = theta

            cam1_block = True
            cam2_block = True
            if cam1_block: # for readability
                cam1_x_values = np.concatenate([
                    x3 + centers[0], 
                    [centers[0], centers[-1]],
                    x3_b + centers[-1],
                ])
                cam1_y_values = np.concatenate([
                    y3, 
                    [-self.R3, -self.R3],
                    y3_b, 
                ])
                self.ax.plot(cam1_x_values, cam1_y_values, color='black', linestyle='-') # cam profile 1

                x_line = [-self.bar_length/2+self.x, self.bar_length/2+self.x]
                y_line = [-self.R1*self.previous_direction, -self.R1*self.previous_direction]
                y_line_floating = [-self.R1*self.previous_direction-self.bar_height_max_disp* self.previous_direction, -self.R1*self.previous_direction-self.bar_height_max_disp* self.previous_direction]

                # Calculate midpoint
                midpoint_x = (x_line[0] + x_line[1]) / 2
                midpoint_y = (y_line[0] + y_line[1]) / 2

                # Length of the new line
                length = self.R4

                x_new_line = [midpoint_x, midpoint_x]
                y_new_line =          [midpoint_y , midpoint_y - length *self.previous_direction]
                y_new_line_floating = [midpoint_y -self.bar_height_max_disp* self.previous_direction, midpoint_y - length *self.previous_direction-self.bar_height_max_disp* self.previous_direction]

                x_line_floating = x_line.copy()
                x_new_line_floating = x_new_line.copy()
                cam1_active, midpoint, closest_point, [x_line, y_line],[x_line_floating,y_line_floating], [x_new_line,y_new_line],[x_new_line_floating, y_new_line_floating]= self.cam_is_active(cam1_x_values, cam1_y_values, [midpoint_x, midpoint_y],[x_line, y_line],[x_line_floating,y_line_floating], [x_new_line,y_new_line], [x_new_line_floating, y_new_line_floating],self.R3 - self.R1)
                if cam1_active:
                    # ____1
                    self.ax.plot([midpoint[0], closest_point[0]], [midpoint[1], closest_point[1]], color='green', linestyle='--')
                    self.ax.plot(closest_point[0], closest_point[1], 'ko') 
                    d_x = self.R1 * d_theta

                else:
                    y_line = y_line_floating.copy()
                    x_line = x_line_floating.copy()
                    x_new_line = x_new_line_floating.copy()
                    y_new_line = y_new_line_floating.copy()

                x_1_line = x_line
                y_1_line = y_line
                x_p_1_line = x_new_line
                y_p_1_line = y_new_line


            if cam2_block: # for readability
                cam2_x_values = np.concatenate([
                    x4_b + centers[-1],
                    [centers[0], centers[-1]],
                    x4 + centers[0], 
                ])
                cam2_y_values = np.concatenate([
                    y4_b, 
                    [self.R4, self.R4],
                    y4, 
                ])
                self.ax.plot(cam2_x_values, cam2_y_values, color='red', linestyle='-') # cam profile 2


                x_line = [-self.bar_length/2+self.x, self.bar_length/2+self.x]
                y_line = [-self.R2*self.previous_direction, -self.R2*self.previous_direction]
                y_line_floating = [-self.R2*self.previous_direction-self.bar_height_max_disp* self.previous_direction, -self.R2*self.previous_direction-self.bar_height_max_disp* self.previous_direction]

                # Calculate midpoint
                midpoint_x = (x_line[0] + x_line[1]) / 2
                midpoint_y = (y_line[0] + y_line[1]) / 2

                length_perpendicular = self.R4 

                x_perpendicular = [midpoint_x, midpoint_x]
                y_perpendicular =     [midpoint_y, midpoint_y - length_perpendicular*self.previous_direction]
                y_new_line_floating = [midpoint_y - self.bar_height_max_disp* self.previous_direction, midpoint_y - length_perpendicular*self.previous_direction - self.bar_height_max_disp* self.previous_direction]

                x_line_floating = x_line.copy()
                x_new_line_floating = x_perpendicular.copy()

                cam2_active, midpoint, closest_point, [x_line, y_line], [x_line_floating,y_line_floating], [x_perpendicular, y_perpendicular],[x_new_line_floating, y_new_line_floating]= self.cam_is_active(cam2_x_values, cam2_y_values, [midpoint_x, midpoint_y], [x_line, y_line],[x_line_floating,y_line_floating], [x_perpendicular,y_perpendicular],[x_new_line_floating, y_new_line_floating], self.R4 - self.R2)
                if cam2_active:
                    self.ax.plot([midpoint[0], closest_point[0]], [midpoint[1], closest_point[1]], color='blue', linestyle='--')
                    self.ax.plot(closest_point[0], closest_point[1], 'ro') 
                    d_x = self.R2 * d_theta
                else:
                    y_line = y_line_floating.copy()
                    x_line = x_line_floating.copy()
                    x_perpendicular = x_new_line_floating
                    y_perpendicular = y_new_line_floating

                x_2_line = x_line
                y_2_line = y_line
                x_p_2_line = x_perpendicular
                y_p_2_line = y_perpendicular
                    

            # we can handle the previous direction here....
            if self.direction != self.previous_direction:
                self.previous_direction = self.direction

            mod_arm_angle = self.arm_angle# % (-2*np.pi)
            if self.x == centers[0]:
                if -np.pi > self.arm_angle and self.arm_angle > -2*np.pi:
                    self.arm_angle = - np.pi
                elif self.arm_angle > 0:
                    self.arm_angle = 0
                elif self.arm_angle <= -2*np.pi:
                    self.arm_angle = 0
                mod_arm_angle = self.arm_angle
                if (mod_arm_angle) <= -np.pi:
                    self.direction = -1
                elif mod_arm_angle >= 0:
                    self.direction = 1
            if self.x == centers[-1]:
                if 0 > self.arm_angle and self.arm_angle > -np.pi:
                    self.arm_angle = -np.pi
                elif self.arm_angle < -2*np.pi:
                    self.arm_angle = -2*np.pi
                elif self.arm_angle >= 0:
                    self.arm_angle = -2*np.pi
                mod_arm_angle = self.arm_angle
                if (mod_arm_angle) >= -np.pi:
                    self.direction = -1
                elif (mod_arm_angle) <= -2*np.pi:
                    self.direction = 1
            
            d_x *= self.direction
            if self.x-d_x < centers[0] or (0 > self.arm_angle and self.arm_angle > -np.pi): # arm is rotating
                self.x = centers[0]
                self.arm_angle -= d_theta

                if -np.pi > self.arm_angle and self.arm_angle > -2*np.pi:
                    self.arm_angle = - np.pi
                elif self.arm_angle > 0:
                    self.arm_angle = 0
                elif self.arm_angle <= -2*np.pi:
                    self.arm_angle = 0
                
            elif self.x-d_x > centers[-1] or (-np.pi > self.arm_angle and self.arm_angle > -2*np.pi): # arm is rotating
                self.x = centers[-1]
                self.arm_angle -= d_theta

                if 0 > self.arm_angle and self.arm_angle > -np.pi:
                    self.arm_angle = -np.pi
                elif self.arm_angle < -2*np.pi:
                    self.arm_angle = -2*np.pi
                elif self.arm_angle >= 0:
                    self.arm_angle = -2*np.pi

            else:
                self.x-=d_x
            
            self.ax.plot(x_1_line, y_1_line, color='black', linestyle='-', linewidth=3)
            self.ax.plot(x_p_1_line, y_p_1_line, color='black', linestyle='-')
            self.ax.plot(x_2_line, y_2_line, color='red', linestyle='-', linewidth=3)
            self.ax.plot(x_p_2_line, y_p_2_line, color='red', linestyle='-')
            
            
            self.ax.plot(centers, [0] * len(centers), 'ro')
            for key, value in self.__dict__.items():
                if key in instance:  # Check if the key exists in the instance
                    instance[key] = value  # Update the dictionary
                elif hasattr(instance, '__dict__') and key in vars(instance):
                    setattr(instance, key, value)
        return self

    def update(self, interval=100, frames=100):
        speed = self.settings["speed"]
        len = self.settings["animation_rounds"]
        displacements = np.linspace(0, len, int(250/speed))
        displacements = np.concatenate([displacements, displacements[::-1]])  # Reverse motion
        def animate(displacement):
            self.ax.clear()
            self.plot(displacement)

        
        self.setup_plot()
        self.ani = animation.FuncAnimation(self.fig, lambda displacement: animate(displacement), frames=displacements, interval=50, repeat=True)
        return self

    def show(self):
        plt.show()
        return self

    def save(self, fileName):
        if self.ani is not None:
            # writer = PillowWriter(fps=25)
            self.ani.save(f"{fileName}.gif", writer='PillowWriter') # how to set fps
        else:
            plt.savefig(f"{fileName}.png")
        return self

    @staticmethod
    def Settings(R1=1,R2=4.5,R3=9, R4=10, L=10, num_wheels=3, num_instances = 2, speed=0.25, animation_rounds=5):
        
        centers = [i * L for i in range(num_wheels)]
        # center the centers
        origin = centers[-1]/2
        centers[:] = [center - origin for center in centers] 
        x_limits = [centers[0]-2*R4,centers[-1]+2*R4]

        # instance params
        # the
            # self.x = 0
            # self.previous_theta = 0
            # self.direction = 1
            # self.previous_direction = 1
            # self.arm_angle = 0
            # self.theta_0 = 0
        ## calculate theta to complete the cirlce
        ## divide equally by the number of instances
        perimeter_theta = ((centers[-1] - centers[0] ))/R2 + ((centers[-1] - centers[0] )/R1) + (2*np.pi)
        theta_points = [(perimeter_theta/num_instances) * i for i in range(num_instances)]
        # calculate the corresponding xes
        theta_section1 = (((centers[-1] - centers[0] ))/2)/R1
        theta_section2 = theta_section1 + np.pi
        theta_section3 = theta_section2 + ((centers[-1] - centers[0] ))/R2
        theta_section4 = theta_section3 + np.pi
        theta_section5 = theta_section4 + (((centers[-1] - centers[0] ))/2)/R1
        theta_sections = [theta_section1, theta_section2, theta_section3, theta_section4, theta_section5]
        x_limits1 = -(centers[-1] - centers[0] ) /2 
        x_limits2 = x_limits1 
        x_limits3 = (centers[-1] - centers[0] )/2
        x_limits4 = x_limits3
        x_limits5 = 0

        animation_rounds = int(animation_rounds * perimeter_theta)

        def get_x(theta): # [x, direction, theta, arm_angle]
            theta = theta % perimeter_theta
            if 0 <=theta and theta < theta_sections[0]:
                return [0 - theta * R1, 1, -theta, 0] 
            elif theta_sections[0] <= theta and theta < theta_sections[1]:
                return [x_limits1,1 , -theta, -(theta-theta_sections[0])]
            elif theta_sections[1] <theta and theta < theta_sections[2]:
                return [x_limits2 + (theta - theta_section2) * R2 ,-1, -theta, -np.pi]
            elif theta_sections[2] <theta and theta < theta_sections[3]:
                return [x_limits3,-1, -theta, -np.pi - (theta-theta_section3)]
            else:
                return [x_limits3 - (theta - theta_section4) * R1,1, -theta, -2*np.pi]
            
        starting_xes = [get_x(theta) for theta in theta_points]
        instances = [{"x":instance[0], "direction":instance[1], "theta":instance[2], "arm_angle":instance[3],} for instance in starting_xes]
        for instance in instances:
            instance["previous_direction"] = instance["direction"]
            instance["previous_theta"] = instance["theta"]
            instance["theta_offset"] = instance["theta"]
        # print(instances)

        return {
            "R1": R1,
            "R2": R2,
            "R3": R3,
            "R4": R4,
            "L": L,
            "centers": centers,
            "origin":origin,
            "x_limits": x_limits,
            "instances": instances,
            "animation_rounds": animation_rounds,
            "speed": speed,
        }
    @staticmethod
    def info():
        description = (
            "Mechanism Description:\n"
            "-----------------------\n"
            "Multi-legged ULMW with quick return.\n"
            "**Applications**:\n"
            "This type of mechanism is intended to be used in the development of clamp-on self-actuated long-range linear motion mechanisms.\n"
        )
        return description


import numpy as np
import matplotlib.pyplot as plt

class bmutils:
    # def is_point_on_circle(self, R_intersect, x_B, y_B, tolerance=0.1, index=0):
    #     """
    #     Check if R_intersect falls on the circle defined by points (x_B, y_B).
        
    #     Parameters:
    #     R_intersect (list): The point to check, [x, y].
    #     x_B (array-like): X coordinates of the circle points.
    #     y_B (array-like): Y coordinates of the circle points.
    #     tolerance (float): Distance tolerance to consider for "on the circle".
        
    #     Returns:
    #     bool: True if R_intersect is on the circle, False otherwise.
    #     """
    #     for x, y in zip(x_B, y_B):
    #         distance = np.sqrt((R_intersect[0] - x) ** 2 + (R_intersect[1] - y) ** 2)
    #         # print(distance, index)
    #         if distance < tolerance:
    #             return True
    #     return False
    # def is_point_on_circle(self, point, x_arc, y_arc, tolerance=1e-5, index=0):
    def is_point_on_circle(self, point, x_arc, y_arc, tolerance=0.1):
        # Initialize variables to track the closest distance and point
        closest_distance = float('inf')
        closest_point = None

        # Iterate through pairs of points to find the closest segment
        for i in range(len(x_arc) - 1):
            # Get the current and next point
            p1 = (x_arc[i], y_arc[i])
            p2 = (x_arc[i + 1], y_arc[i + 1])
            
            # Project the point onto the line segment
            closest_on_segment = self.closest_point_on_segment(point, p1, p2)
            
            # Calculate the distance to this closest point
            distance = np.sqrt((closest_on_segment[0] - point[0])**2 + (closest_on_segment[1] - point[1])**2)
            
            # Update if this is the closest point so far
            if distance < closest_distance:
                closest_distance = distance
                closest_point = closest_on_segment

        # # Plot the closest point
        # if closest_distance <= tolerance:
        #     if closest_point is not None:
        #         self.ax.plot(closest_point[0], closest_point[1], 'go')  # __go__

        # Check if the closest distance is within the specified tolerance
        return closest_distance <= tolerance
    
    def closest_point_on_segment(self, point, p1, p2):
        # Vector from p1 to p2
        segment_vector = np.array(p2) - np.array(p1)
        # Vector from p1 to the point
        point_vector = np.array(point) - np.array(p1)

        # Project point_vector onto segment_vector
        segment_length_squared = np.dot(segment_vector, segment_vector)
        if segment_length_squared == 0:  # p1 and p2 are the same point
            return p1

        projection = np.dot(point_vector, segment_vector) / segment_length_squared
        projection = np.clip(projection, 0, 1)  # Clamp to the segment
        closest_point = p1 + projection * segment_vector

        return closest_point

    def watermark(self, ax):
        ax.text(0.5, 0.5, 'BrianMechanisms', transform=ax.transAxes,
                fontsize=40, color='gray', alpha=0.5,
                ha='center', va='center', rotation=30)
        
    def create_circle_section(self, r, num_points=100, start_angle=0, end_angle=2 * np.pi, phase=0):
        theta = np.linspace(start_angle + phase, end_angle + phase, num_points)
        x = r * np.cos(theta)
        y = r * np.sin(theta)
        return x, y
    
    def rotate_points(self, theta_radians, x_points, y_points , center):
        center_x, center_y = center
        # x_points, y_points = points
        rotation_matrix = np.array([[np.cos(theta_radians), -np.sin(theta_radians)],
                                    [np.sin(theta_radians), np.cos(theta_radians)]])
        # Convert lists to numpy arrays
        x_points = np.array(x_points)
        y_points = np.array(y_points)

        # Translate points to the origin
        translated_x = x_points - center_x
        translated_y = y_points - center_y
        
        # Rotate points
        rotated = rotation_matrix @ np.array([translated_x, translated_y])
        
        # Translate points back
        return rotated[0] + center_x, rotated[1] + center_y
    
    def show(self):
        plt.show()
        return self
    
    def save(self, fileName):
        if self.ani is not None:
            # writer = PillowWriter(fps=25)
            self.ani.save(f"{fileName}.gif", writer='PillowWriter') # how to set fps
        else:
            plt.savefig(f"{fileName}.png")
        return self


import sys
import os
import numpy as np


import numpy as np
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
from ....utils import bmutils

class mechanism33(bmutils):
    def __init__(self, settings):
        self.settings = settings
        self.fig, self.ax = plt.subplots(figsize=(10, 6))
        self.ani = None

        # Extract parameters from settings
        self.start_theta = settings['start_theta']
        self.radius = settings['radius']
        self.instances = settings['instances']
        self.forward_stroke_angle = settings['forward_stroke_angle']
        self.stroke_length = settings['stroke_length']
        # self.b = settings['b']

        # Initialize plot elements
        self.point_line = [] # = self.ax.plot([], [], color='green', linestyle='-', linewidth=2, label='Line to Point A')
        self.point_dot = [] # self.ax.scatter([], [], color='red', zorder=5)
        self.center_dot = self.ax.scatter([0], [0], color='red', zorder=5)  # Fixed here (no need to unpack)
        # self.pointB_dot = self.ax.scatter([], [], color='red', zorder=5)
        # self.point_label = self.ax.text(0, 0, 'A', fontsize=12, ha='right', color='red')

        self.cam_profile_line = []
        self.return_line = []
        for instance in self.instances:
            point_line, = self.ax.plot([], [], color=instance["color"], linestyle='-', linewidth=2, label='Crank')
            point_dot = self.ax.scatter([], [], color=instance["color"], zorder=5)

            self.point_dot.append(point_dot)
            self.point_line.append(point_line)
            cam_profile_line, = self.ax.plot([], [], color=instance["color"], linestyle='-', linewidth=2)
            return_line, = self.ax.plot([], [], color=instance["color"], linestyle='-', linewidth=2)
            self.cam_profile_line.append(cam_profile_line)
            self.return_line.append(return_line)

        self.setup_plot()

    def setup_plot(self):
        self.ax.set_title("Mechanism 33")
        self.ax.set_aspect('equal')
        self.ax.grid(True)
        self.ax.axis("off")
        radius = self.radius
        self.ax.set_xlim(-2 * radius, 2 * radius)
        self.ax.set_ylim(-1.2 * radius, 1.2 * radius)
        self.follower_circle()
        super().watermark(self.ax)

    def follower_circle(self):
        theta_circle = np.linspace(0, 2*np.pi, 1000)
        x_circle = self.radius * np.cos(theta_circle)
        y_circle = self.radius * np.sin(theta_circle)
        self.ax.plot(x_circle, y_circle, label='Circle', color='blue')

    def get_cam_profile(self, theta, d_theta):
        start_theta = self.start_theta
        displacement_per_theta = self.stroke_length / self.forward_stroke_angle
        d_theta = d_theta % (2*np.pi)

        theta_rad = np.linspace(start_theta, start_theta + np.deg2rad(self.forward_stroke_angle), 1000)
        x, y = self.radius * np.cos(theta_rad), self.radius * np.sin(theta_rad)

        d_theta_profile = np.rad2deg(theta_rad - start_theta)

        point_x =       self.radius * np.cos(theta)
        x_for_profile = point_x + (np.rad2deg(d_theta)) * displacement_per_theta

        x_diff = x_for_profile - point_x         
        
        if 0 <= d_theta and d_theta <= np.deg2rad(self.forward_stroke_angle):
            x_profile_displaced = (x + d_theta_profile * displacement_per_theta - x_diff)
        else:
            x_profile_displaced = (x + d_theta_profile * displacement_per_theta)
            x1, x2 = x_profile_displaced[0], x_profile_displaced[-1]
            y1, y2 = y[0], y[-1]
            m = (y2 - y1) / (x2 - x1)
            b = y1 - m * x1
            y_target = self.radius * np.sin(theta)
            x_for_y = (y_target - b) / m
            x_diff = x_for_y - point_x
            x_profile_displaced = (x + d_theta_profile * displacement_per_theta - x_diff)
   
        return x_profile_displaced, y


    def plot(self, d_theta=0):
        theta = self.start_theta + d_theta
        for i, instance in enumerate(self.instances):
            instance_theta = theta + np.deg2rad(instance["offset"])
            instance_d_theta = d_theta + np.deg2rad(instance["offset"])
            point_x = self.radius * np.cos(instance_theta)
            point_y = self.radius * np.sin(instance_theta)
            x_profile_displaced, y = self.get_cam_profile(instance_theta, instance_d_theta)
            self.cam_profile_line[i].set_data(x_profile_displaced, y)
            self.return_line[i].set_data([x_profile_displaced[0], x_profile_displaced[-1]], [y[0], y[-1]])
            self.point_line[i].set_data([0, point_x], [0, point_y])
            self.point_dot[i].set_offsets([point_x, point_y])

        return self

    def update(self, interval=50, frames=100):
        speed = self.settings.get("speed", 1)
        animation_rounds = self.settings.get("animation_rounds", 1)
        total_angle = 2*np.pi * animation_rounds
        displacements = np.linspace(0, total_angle, int(frames/speed))

        def animate(displacement):
            self.plot(displacement)

        self.ani = FuncAnimation(self.fig, lambda displacement: animate(displacement),
                                 frames=displacements, interval=interval, repeat=True)
        return self

    @staticmethod
    def Settings(radius=1, forward_stroke_angle=250, speed=1, animation_rounds=1, num_instances = 2, stroke_length=1.2, start_theta=-160):

        colors = [
           "black","orange", "green", "gray"
        ]
        instances = [{"color":colors[i%len(colors)], "offset": 360/num_instances *( i+0)} for i in range(num_instances)]
        return {
            "radius": radius,
            "forward_stroke_angle": forward_stroke_angle,
            "stroke_length": stroke_length,
            # "b": b,
            "speed": speed,
            "animation_rounds": animation_rounds,
            "instances": instances,
            "start_theta": np.deg2rad(start_theta)#-np.pi/2
        }

    @staticmethod
    def info():
        return (
            "Mechanism Description:\n"
            "-----------------------\n"
            "A variation of scotch-yoke with constant speed in forward stroke and quickreturn.\n"
            "**Applications**:\n"
            "Clamp-on long range linear motion mechanisms and walking mechanisms\n"
            "**Notes**:\n"
            "You may need to select values for stroke_length, start_theta and  forward_stroke_angle for which the cam profile will not cross the shaft axis in the case where multiple co-axial cams are used\n"
        )


from .locii import *
from .appproximateStraightLineMechanisms import *
from .mirror import *
from .straightLineMechanisms import *
from .helpers import *
from .ULMWGripWalker import *
from .leggedTracks import *
from .leggedTracksTrue import *
from .reciprocating import *


import numpy as np
import re
import copy
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
import math

class Locii:
    def __init__(self, links, configs):
        self.links = links
        self.configs = configs

    def getPivot(self, link):
        for key, value in link.get("points").items():
            if value == 0:
                return key

    def getOrigin(self, link):
        origin = link.get("origin")
        if origin is not None and origin[0] == "{" and origin[-1] == "}":
            return tuple(map(float, origin[1:-1].split(',')))
        else:
            link_name, point = origin.split(".")
            return self.links.get(link_name).get("positions").get(point)

    def extract_variables(self, equation):
        pattern = re.compile(r'\b[a-zA-Z_][a-zA-Z0-9_]*\b')
        matches = pattern.findall(equation)
        functions = {"np", "sin", "cos", "tan", "log", "exp", "sqrt"}
        variables = [match for match in matches if match not in functions]
        return variables

    def replace_variables(self, equation, replacements):
        def replacer(match):
            var = match.group(0)
            return replacements.get(var, var)
        pattern = re.compile(r'\b[a-zA-Z_][a-zA-Z0-9_]*\b')
        new_equation = pattern.sub(replacer, equation)
        return new_equation

    def evaluateKeys(self, link, keys_to_evaluate):
        for key in keys_to_evaluate:
            if key in link:
                value = link[key]
                if isinstance(value, str):
                    if value[0:4] == "link":
                        link_name, link_key = value.split(".")
                        link[key] = self.links.get(link_name).get(f"calculated_{link_key}")
                    else:
                        equation_variables = self.extract_variables(value)
                        replacements = {}
                        for variable in equation_variables:
                            if variable[0:4] == "link":
                                link_name, link_key = variable.split(".")
                                replacements[variable] = str(self.links.get(link_name).get(f"calculated_{link_key}"))
                            else:
                                replace = link.get(f"calculated_{variable}") if f"calculated_{variable}" in link else link.get(variable) if variable in link else None
                                if replace is not None:
                                    replacements[variable] = f"{replace}"
                        link[key] = self.replace_variables(value, replacements)
                else:
                    link[f"calculated_{key}"] = value
        return link

    def getLowestStepDifference(self):
        links = self.links
        second_points = []
        for link in links:            
            if list(links[link]["positions"].values())[0] != (0, 0):
                second_point = list(links[link]["positions"].values())[1]
                second_points.append(second_point)
        y_values = [point[1] for point in second_points]
        sorted_y_values = sorted(y_values)
        lowest_y = sorted_y_values[0]
        second_lowest_y = sorted_y_values[1]
        difference = second_lowest_y - lowest_y
        return lowest_y, second_lowest_y, difference
    
    def plotStepHeight(self):
        fig, axs = plt.subplots(1, 2)
        step_heights = []
        lowest_ys = []
        theta_range = np.arange(0, 361, 1)
        for theta in theta_range:
            paths = self.outputPath(np.arange(theta, theta + 1, 1))
            lowest_y, _, difference = self.getLowestStepDifference()
            lowest_ys.append(lowest_y)
        lowest_y = min(lowest_ys)

        actual_steps = []
        for theta in theta_range:
            paths = self.outputPath(np.arange(theta, theta + 1, 1))
            paths = paths[0]
            lowest_y_for_step, _, difference = self.getLowestStepDifference()
            actual_step = lowest_y_for_step - lowest_y
            actual_steps.append(actual_step)
            step_heights.append(difference)


        axs[0].set_title('Diff btn 2 lowest legs')
        axs[0].plot(np.arange(0, 361, 1), step_heights)
        axs[0].set_xlabel('Theta')
        axs[0].set_ylabel('Step Height')
        # mark on the plot the x and y value of the step_heights
        min_step_height = min(step_heights)
        max_step_height = max(step_heights)
        min_theta = theta_range[step_heights.index(min_step_height)]
        max_theta = theta_range[step_heights.index(max_step_height)]

        lowest_y_at_lowest_step = lowest_ys[step_heights.index(min_step_height)]
        step_height_at_lowest_y = lowest_y_at_lowest_step - lowest_y

        axs[0].annotate(f'Min: {min_step_height:.2f}, {min_theta:.2f}', xy=(min_theta, min_step_height), 
                    xytext=(min_theta, min_step_height + 0.05),
                    arrowprops=dict(facecolor='black', shrink=0.05),
                    horizontalalignment='center')
        
        axs[0].annotate(f'Max: {max_step_height:.2f}', xy=(max_theta, max_step_height), 
                    xytext=(max_theta, max_step_height - 0.1),
                    arrowprops=dict(facecolor='black', shrink=0.05),
                    horizontalalignment='center')
        
        # plot also actual steps
        axs[1].set_title('Step Height')
        axs[1].plot(np.arange(0, 361, 1), actual_steps)
        axs[1].set_xlabel('Theta')
        axs[1].set_ylabel('Actual Step Height')
        
        min_actual_step = min(actual_steps)
        max_actual_step = max(actual_steps)
        min_theta = theta_range[actual_steps.index(min_actual_step)]
        max_theta = theta_range[actual_steps.index(max_actual_step)]

        axs[1].annotate(f'Min: {min_actual_step:.2f}, {min_theta:.2f}', xy=(min_theta, min_actual_step),
                    xytext=(min_theta, min_actual_step + 0.002),
                    arrowprops=dict(facecolor='black', shrink=0.05),
                    horizontalalignment='center')
        
        axs[1].annotate(f'Max: {max_actual_step:.5f}', xy=(max_theta, max_actual_step),
                    xytext=(max_theta, max_actual_step - 0.002),
                    arrowprops=dict(facecolor='black', shrink=0.05),
                    horizontalalignment='center')
        
        return fig, axs

    def rotate_point(self, point, theta0):
        x, y = point
        theta = np.radians(theta0)  # Convert angle from degrees to radians
        x_prime = x * np.cos(theta) - y * np.sin(theta)
        y_prime = x * np.sin(theta) + y * np.cos(theta)
        return x_prime, y_prime

    def calculatePositions(self, link, theta, linkConfigs={}):
        for key, value in linkConfigs.items():
            link[key] = value
            if isinstance(value, str):
                if value[0:4] == "link":
                    link_name, link_key = value.split(".")
                    link[key] = self.links.get(link_name).get(link_key)
                else:
                    link[key] = eval(value)
            else:
                link[key] = value

        keys_to_evaluate = ["speed", "speed_factor", "speed_source", "equation"]
        link = self.evaluateKeys(link, keys_to_evaluate)

        link["calculated_origin"] = self.getOrigin(link)
        theta0 = link.get("theta0") if "theta0" in link else 0
        theta = np.deg2rad(theta)
        theta0 = np.deg2rad(theta0)
        thetaRotation = self.configs.get("thetaRotation") if "thetaRotation" in self.configs else 0
        positions = {}
        for key in link.get("points").keys():
            positions[key] = (None, None)
        pivot = self.getPivot(link)
        equation = link.get("equation")
        
        points = link.get("points")
        len_factor = link.get("len_factor") if "len_factor" in link else 1
        if isinstance(len_factor, str):
            len_factor = eval(len_factor)

        for pointName, length in points.items():
            length = float(length)
            if equation is None:
                for other_link in self.links.values():
                    if other_link.get("positions") and pointName in other_link["positions"]:
                        point = other_link["positions"][pointName]
                        break
                    else:
                        point = (None, None)  # If 
                # print(pointName)
            else:
                if pointName == pivot:
                    point = link.get("calculated_origin") 
                else:
                    x_eq, y_eq = equation.split(',')
                    unrotatedOrigin = link.get("calculated_origin") # do not rotate a point that has been rotated
                    unrotatedOrigin = self.rotate_point(unrotatedOrigin, -thetaRotation) 
                    # print(x_eq, y_eq)
                    x_position = len_factor * length * eval(x_eq.replace("x", str(theta0 + theta))) + unrotatedOrigin[0]
                    y_position = len_factor * length * eval(y_eq.replace("x", str(theta0 + theta))) + unrotatedOrigin[1]
                    point = (x_position, y_position)
                    # if link.get("rotate_by_mechanism_theta"):
                    #     pass 
                    # else:
                    point = self.rotate_point(point, thetaRotation) # Rotate the point
            positions[pointName] =  point#(x_position, y_position) #
            # if link.get("rotate_by_mechanism_theta"):
            #     print("rotating by mechanism theta")
                
        link["positions"] = positions
        return link

    def outputPath(self, thetaRange=None):
        if thetaRange is None:
            thetaRange = np.arange(0, self.configs.get("thetaMax", 360)+1, 1)
        paths = []
        for theta in thetaRange:
            thetaPoints = {}
            for key, link in self.links.items():
                link = self.calculatePositions(link, theta, self.configs[key] if key in self.configs else {})
                thetaPoints[key] = link["positions"]
            paths.append(thetaPoints)
        return paths

    def plotOutPutPaths(self, title={"title": "Output Paths of Links", "sub": None}, plotConfig={"fig": None, "ax": None, "legend": True, "axes": True}):
        fig = plotConfig["fig"]
        ax = plotConfig["ax"]
        if fig is None:
            fig, ax = plt.subplots()
        paths = self.outputPath()
        for link_key, link in self.links.items():
            x_values = {key: [] for key in link["points"].keys()}
            y_values = {key: [] for key in link["points"].keys()}
            output_point = link.get("output")
            pivot = self.getPivot(link)
            for thetaPoints in paths:
                for point, (x, y) in thetaPoints.get(link_key).items():
                    if point != pivot:
                        x_values[point].append(x)
                        y_values[point].append(y)
            for point in x_values.keys():
                if point != pivot:
                    linestyle = 'solid' if point == output_point else 'dotted'
                    ax.plot(x_values[point], y_values[point], label=f'{point}', linestyle=linestyle)

        if ("axes" in plotConfig and plotConfig["axes"]) or "axes" not in plotConfig:
            ax.set_xlabel('X-axis')
            ax.set_ylabel('Y-axis')
        else:
            ax.axis('off')
        fig.suptitle(title["title"], fontsize=16)

        if "sub" in title and title["sub"] is not None:
            ax.set_title(title["sub"])

        ax.set_aspect('equal', adjustable='box')
        if ("legend" in plotConfig and plotConfig["legend"]) or "legend" not in plotConfig:
            ax.legend()

        if "mechanism_theta" in plotConfig:
            mechanism_theta = plotConfig["mechanism_theta"]
            links_mechanism = copy.deepcopy(self.links)
            paths = self.outputPath(np.arange(mechanism_theta, mechanism_theta + 1, 1))
            paths = paths[0]

            all_points = {}
            for link_key, link in paths.items():
                for point, (x, y) in link.items():
                    all_points[point] = (x, y)
            x_coords = [coords[0] for coords in all_points.values()]
            y_coords = [coords[1] for coords in all_points.values()]
            ax.scatter(x_coords, y_coords, color='red')

            for link_key, link in self.links.items():
                points = list(link["points"].keys())
                for i in range(len(points) - 1):
                    x_values = [all_points[points[i]][0], all_points[points[i + 1]][0]]
                    y_values = [all_points[points[i]][1], all_points[points[i + 1]][1]]
                    ax.plot(x_values, y_values, linestyle='solid', label=f'{points[i]} to {points[i + 1]}')

        return fig, ax
    
    def plotVariableLinkLength(self, title={"title": "Length Links", "sub": None}, plotConfig={"fig": None, "ax": None, "legend": True, "axes": True, "axesNames":["x", "y"]}, points={}):
        fig = plotConfig["fig"]
        ax = plotConfig["ax"]
        if fig is None:
            fig, ax = plt.subplots()
        paths = self.outputPath()
        ## the 2 points to compare
        pointA = []
        pointB = []
        pointsList = []
        for link_key, linkPoints in points.items():
            pointsList=points[link_key]
            link = self.links[link_key]
            x_values = {key: [] for key in linkPoints}
            y_values = {key: [] for key in linkPoints}
            
            for thetaPoints in paths:
                for point, (x, y) in thetaPoints.get(link_key).items():
                        x_values[point].append(x)
                        y_values[point].append(y)
                        if point == pointsList[0]:
                            pointA.append([x,y])
                        else:
                            pointB.append([x,y])
           
            thetaRange = np.arange(0, self.configs.get("thetaMax", 360)+1, 1)
            lengths = self.distance_between_points(pointA, pointB)
            ax.plot(thetaRange, lengths, marker='o', linestyle='-', color='b', label='Distances')
                                    
       
        if ("axes" in plotConfig and plotConfig["axes"]) or "axes" not in plotConfig:
            ax.set_xlabel(plotConfig["axesNames"][0])
            ax.set_ylabel(plotConfig["axesNames"][1])
        else:
            ax.axis('off')
        fig.suptitle(title["title"], fontsize=16)

        if "sub" in title and title["sub"] is not None:
            ax.set_title(title["sub"])


        return fig, ax

    def update(self, mechanism_theta, ax, title, plotConfig):
        ax.clear()
        links_mechanism = copy.deepcopy(self.links)
        paths = self.outputPath()
        for link_key, link in self.links.items():
            x_values = {key: [] for key in link["points"].keys()}
            y_values = {key: [] for key in link["points"].keys()}
            output_point = link.get("output")
            pivot = self.getPivot(link)
            for thetaPoints in paths:
                for point, (x, y) in thetaPoints.get(link_key).items():
                    if point != pivot:
                        x_values[point].append(x)
                        y_values[point].append(y)
            for point in x_values.keys():
                if point != pivot:
                    linestyle = 'solid' if point == output_point else 'dotted'
                    ax.plot(x_values[point], y_values[point], label=f'{point}', linestyle=linestyle)

        paths = self.outputPath(np.arange(mechanism_theta, mechanism_theta + 1, 1))
        paths = paths[0]

        all_points = {}
        for link_key, link in paths.items():
            for point, (x, y) in link.items():
                all_points[point] = (x, y)

        x_coords = [coords[0] for coords in all_points.values()]
        y_coords = [coords[1] for coords in all_points.values()]
        ax.scatter(x_coords, y_coords, color='red')

        for link_key, link in self.links.items():
            points = list(link["points"].keys())
            for i in range(len(points) - 1):
                x_values = [all_points[points[i]][0], all_points[points[i + 1]][0]]
                y_values = [all_points[points[i]][1], all_points[points[i + 1]][1]]
                ax.plot(x_values, y_values, linestyle='solid', label=f'{points[i]} to {points[i + 1]}')

        if ("axes" in plotConfig and plotConfig["axes"]) or "axes" not in plotConfig:
            ax.set_xlabel('X-axis')
            ax.set_ylabel('Y-axis')
        else:
            ax.axis('off')

        if "sub" in title and title["sub"] is not None:
            ax.set_title(title["sub"])

        ax.set_aspect('equal', adjustable='box')
        if ("legend" in plotConfig and plotConfig["legend"]) or "legend" not in plotConfig:
            ax.legend()

    def clearPath(self, pathName):
        self.paths.pop(pathName, None)

    def savePath(self, pathName, path):
        self.paths[pathName] = path

    def distance_between_points(self, points1, points2):
        distances = []
        for p1, p2 in zip(points1, points2):
            distance = math.sqrt((p2[0] - p1[0]) ** 2 + (p2[1] - p1[1]) ** 2)
            distances.append(distance)
        return distances
    
    # def distance_between_points(x1, y1, x2, y2):
    #     return math.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2)

    def distance_between_points_3d(x1, y1, z1, x2, y2, z2):
        return math.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2 + (z2 - z1) ** 2)
